# -*- coding: utf-8 -*-
# Copyright (c) 2015, imageio contributors
# imageio is distributed under the terms of the (new) BSD License.

""" Storage of image data in tiff format.
"""

from __future__ import absolute_import, print_function, division

from .. import formats
from ..core import Format, has_module

_itk = None  # Defer loading to load_lib() function.


def load_lib():
    global _itk
    try:
        import SimpleITK as _itk
    except ImportError:
        raise ImportError("SimpleITK could not be found. "
                          "Please try "
                          "  easy_install SimpleITK "
                          "or refer to "
                          "  http://simpleitk.org/ "
                          "for further instructions.")
    return _itk


# Split up in real ITK and all supported formats.
ITK_FORMATS = ('.gipl', '.ipl', '.mha', '.mhd', '.nhdr', '.nii', '.nrrd',
               '.vtk')
ALL_FORMATS = ITK_FORMATS + ('.bmp', '.jpeg', '.jpg', '.png', '.tiff', '.tif',
                             '.dicom', '.gdcm')


class ItkFormat(Format):
    """ The ItkFormat uses the simpleITK library to support a range of
    ITK-related formats. It also supports a few common formats that are
    also supported by the freeimage plugin (e.g. PNG and JPEG).
    
    This format requires the ``simpleITK`` package.

    Parameters for reading
    ----------------------
    None.

    Parameters for saving
    ---------------------
    None.

    """

    def _can_read(self, request):
        # If the request is a format that only this plugin can handle,
        # we report that we can do it; a useful error will be raised
        # when simpleitk is not installed. For the more common formats
        # we only report that we can read if the library is installed.
        if request.filename.lower().endswith(ITK_FORMATS):
            return True
        if has_module('SimpleITK'):
            return request.filename.lower().endswith(ALL_FORMATS)

    def _can_write(self, request):
        if request.filename.lower().endswith(ITK_FORMATS):
            return True
        if has_module('SimpleITK'):
            return request.filename.lower().endswith(ALL_FORMATS)

    # -- reader

    class Reader(Format.Reader):

        def _open(self, **kwargs):
            if not _itk:
                load_lib()
            self._img = _itk.ReadImage(self.request.get_local_filename())

        def _get_length(self):
            return 1

        def _close(self):
            pass

        def _get_data(self, index):
            # Get data
            if index != 0:
                raise IndexError(
                    'Index out of range while reading from itk file')

            # Return array and empty meta data
            return _itk.GetArrayFromImage(self._img), {}

        def _get_meta_data(self, index):
            raise RuntimeError('The itk format does not support '
                               ' meta data.')

    # -- writer
    class Writer(Format.Writer):

        def _open(self):
            if not _itk:
                load_lib()

        def _close(self):
            pass

        def _append_data(self, im, meta):
            _itk_img = _itk.GetImageFromArray(im, isVector=True)
            _itk.WriteImage(_itk_img, self.request.get_local_filename())

        def set_meta_data(self, meta):
            raise RuntimeError('The itk format does not support '
                               ' meta data.')


# Register
title = "Insight Segmentation and Registration Toolkit (ITK) format"
format = ItkFormat('itk', title, ' '.join(ALL_FORMATS), 'iIvV')
formats.add_format(format)



# -*- coding: utf-8 -*-
# Copyright (c) 2015, imageio contributors
# imageio is distributed under the terms of the (new) BSD License.

""" 
These functions represent imageio's main interface for the user. They
provide a common API to read and write image data for a large
variety of formats. All read and write functions accept keyword
arguments, which are passed on to the format that does the actual work.
To see what keyword arguments are supported by a specific format, use
the :func:`.help` function.

Functions for reading:

  * :func:`.imread` - read an image from the specified uri
  * :func:`.mimread` - read a series of images from the specified uri
  * :func:`.volread` - read a volume from the specified uri
  * :func:`.mvolread` - read a series of volumes from the specified uri

Functions for saving:

  * :func:`.imwrite` - write an image to the specified uri
  * :func:`.mimwrite` - write a series of images to the specified uri
  * :func:`.volwrite` - write a volume to the specified uri
  * :func:`.mvolwrite` - write a series of volumes to the specified uri

More control:

For a larger degree of control, imageio provides functions
:func:`.get_reader` and :func:`.get_writer`. They respectively return an
:class:`.Reader` and an :class:`.Writer` object, which can
be used to read/write data and meta data in a more controlled manner.
This also allows specific scientific formats to be exposed in a way
that best suits that file-format.

----

Supported resource URI's:

All functions described here accept a URI to describe the resource to
read from or write to. These can be a wide range of things. (Imageio
takes care of handling the URI so that plugins can access the data in
an easy way.)

For reading and writing:

* a normal filename, e.g. ``'c:\\foo\\bar.png'``
* a file in a zipfile, e.g. ``'c:\\foo\\bar.zip\\eggs.png'``
* a file object with a ``read()`` / ``write()`` method.

For reading:

* an http/ftp address, e.g. ``'http://example.com/foo.png'``
* the raw bytes of an image file
* ``get_reader("<video0>")`` to grab images from a (web) camera.
* ``imread("<screen>")`` to grab a screenshot (on Windows or OS X).
* ``imread("<clipboard>")`` to grab an image from the clipboard (on Windows).

For writing one can also use ``'<bytes>'`` or ``imageio.RETURN_BYTES`` to
make a write function return the bytes instead of writing to a file.

"""

from __future__ import absolute_import, print_function, division

import numpy as np

from . import Request, RETURN_BYTES
from .. import formats


def help(name=None):
    """ help(name=None)
    
    Print the documentation of the format specified by name, or a list
    of supported formats if name is omitted. 
    
    Parameters
    ----------
    name : str
        Can be the name of a format, a filename extension, or a full
        filename. See also the :doc:`formats page <formats>`.
    """
    if not name:
        print(formats)
    else:
        print(formats[name])


## Base functions that return a reader/writer


def get_reader(uri, format=None, mode='?', **kwargs):
    """ get_reader(uri, format=None, mode='?', **kwargs)
    
    Returns a :class:`.Reader` object which can be used to read data
    and meta data from the specified file.
    
    Parameters
    ----------
    uri : {str, bytes, file}
        The resource to load the image from, e.g. a filename, http address or
        file object, see the docs for more info.
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename and its contents.
    mode : {'i', 'I', 'v', 'V', '?'}
        Used to give the reader a hint on what the user expects (default "?"):
        "i" for an image, "I" for multiple images, "v" for a volume,
        "V" for multiple volumes, "?" for don't care.
    kwargs : ...
        Further keyword arguments are passed to the reader. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    # Create request object
    request = Request(uri, 'r' + mode, **kwargs)
    
    # Get format
    if format is not None:
        format = formats[format]
    else:
        format = formats.search_read_format(request)
    if format is None:
        raise ValueError('Could not find a format to read the specified file '
                         'in mode %r' % mode)
    
    # Return its reader object
    return format.get_reader(request)


def get_writer(uri, format=None, mode='?', **kwargs):
    """ get_writer(uri, format=None, mode='?', **kwargs)
    
    Returns a :class:`.Writer` object which can be used to write data
    and meta data to the specified file.
    
    Parameters
    ----------
    uri : {str, file}
        The resource to write the image to, e.g. a filename or file object,
        see the docs for more info.
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename.
    mode : {'i', 'I', 'v', 'V', '?'}
        Used to give the writer a hint on what the user expects (default '?'):
        "i" for an image, "I" for multiple images, "v" for a volume,
        "V" for multiple volumes, "?" for don't care.
    kwargs : ...
        Further keyword arguments are passed to the writer. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    # Signal extension when returning as bytes, needed by e.g. ffmpeg
    if uri == RETURN_BYTES and isinstance(format, str):
        uri = RETURN_BYTES + '.' + format.strip('. ')
    
    # Create request object
    request = Request(uri, 'w' + mode, **kwargs)
    
    # Get format
    if format is not None:
        format = formats[format]
    else:
        format = formats.search_write_format(request)
    if format is None:
        raise ValueError('Could not find a format to write the specified file '
                         'in mode %r' % mode)
    
    # Return its writer object
    return format.get_writer(request)


## Images

def imread(uri, format=None, **kwargs):
    """ imread(uri, format=None, **kwargs)
    
    Reads an image from the specified file. Returns a numpy array, which
    comes with a dict of meta data at its 'meta' attribute.
    
    Note that the image data is returned as-is, and may not always have
    a dtype of uint8 (and thus may differ from what e.g. PIL returns).
    
    Parameters
    ----------
    uri : {str, bytes, file}
        The resource to load the image from, e.g. a filename, http address or
        file object, see the docs for more info.
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename and its contents.
    kwargs : ...
        Further keyword arguments are passed to the reader. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    # Get reader and read first
    reader = read(uri, format, 'i', **kwargs)
    with reader:
        return reader.get_data(0)


def imwrite(uri, im, format=None, **kwargs):
    """ imwrite(uri, im, format=None, **kwargs)
    
    Write an image to the specified file.
    
    Parameters
    ----------
    uri : {str, file}
        The resource to write the image to, e.g. a filename or file object,
        see the docs for more info.
    im : numpy.ndarray
        The image data. Must be NxM, NxMx3 or NxMx4.
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename and its contents.
    kwargs : ...
        Further keyword arguments are passed to the writer. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    # Test image
    if isinstance(im, np.ndarray):
        if im.ndim == 2:
            pass
        elif im.ndim == 3 and im.shape[2] in [1, 3, 4]:
            pass
        else:
            raise ValueError('Image must be 2D (grayscale, RGB, or RGBA).')
    else:
        raise ValueError('Image must be a numpy array.')
    
    # Get writer and write first
    writer = get_writer(uri, format, 'i', **kwargs)
    with writer:
        writer.append_data(im)
    
    # Return a result if there is any
    return writer.request.get_result()


## Multiple images

def mimread(uri, format=None, memtest=True, **kwargs):
    """ mimread(uri, format=None, memtest=True, **kwargs)
    
    Reads multiple images from the specified file. Returns a list of
    numpy arrays, each with a dict of meta data at its 'meta' attribute.
    
    Parameters
    ----------
    uri : {str, bytes, file}
        The resource to load the images from, e.g. a filename, http address or
        file object, see the docs for more info.
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename and its contents.
    memtest : bool
        If True (default), this function will raise an error if the
        resulting list of images consumes over 256 MB of memory. This
        is to protect the system using so much memory that it needs to
        resort to swapping, and thereby stall the computer. E.g.
        ``mimread('hunger_games.avi')``.
    kwargs : ...
        Further keyword arguments are passed to the reader. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    # Get reader
    reader = read(uri, format, 'I', **kwargs)
    
    # Read
    ims = []
    nbytes = 0
    for im in reader:
        ims.append(im)
        # Memory check
        nbytes += im.nbytes
        if memtest and nbytes > 256 * 1024 * 1024:
            ims[:] = []  # clear to free the memory
            raise RuntimeError('imageio.mimread() has read over 256 MiB of '
                               'image data.\nStopped to avoid memory problems.'
                               ' Use imageio.get_reader() or memtest=False.')
    
    return ims


def mimwrite(uri, ims, format=None, **kwargs):
    """ mimwrite(uri, ims, format=None, **kwargs)
    
    Write multiple images to the specified file.
    
    Parameters
    ----------
    uri : {str, file}
        The resource to write the images to, e.g. a filename or file object,
        see the docs for more info.
    ims : sequence of numpy arrays
        The image data. Each array must be NxM, NxMx3 or NxMx4.
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename and its contents.
    kwargs : ...
        Further keyword arguments are passed to the writer. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    #Ensure non empty sequence
    if len(ims) == 0:
        raise ValueError('Numpy sequence must not be empty.')
    
    # Get writer
    writer = get_writer(uri, format, 'I', **kwargs)
    with writer:
        
        # Iterate over images (ims may be a generator)
        for im in ims:
            
            # Test image
            if isinstance(im, np.ndarray):
                if im.ndim == 2:
                    pass
                elif im.ndim == 3 and im.shape[2] in [1, 3, 4]:
                    pass
                else:
                    raise ValueError('Image must be 2D '
                                     '(grayscale, RGB, or RGBA).')
            else:
                raise ValueError('Image must be a numpy array.')
            
            # Add image
            writer.append_data(im)
    
    # Return a result if there is any
    return writer.request.get_result()


## Volumes

def volread(uri, format=None, **kwargs):
    """ volread(uri, format=None, **kwargs)
    
    Reads a volume from the specified file. Returns a numpy array, which
    comes with a dict of meta data at its 'meta' attribute.
    
    Parameters
    ----------
    uri : {str, bytes, file}
        The resource to load the volume from, e.g. a filename, http address or
        file object, see the docs for more info.
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename and its contents.
    kwargs : ...
        Further keyword arguments are passed to the reader. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    # Get reader and read first
    reader = read(uri, format, 'v', **kwargs)
    with reader:
        return reader.get_data(0)


def volwrite(uri, im, format=None, **kwargs):
    """ volwrite(uri, vol, format=None, **kwargs)
    
    Write a volume to the specified file.
    
    Parameters
    ----------
    uri : {str, file}
        The resource to write the image to, e.g. a filename or file object,
        see the docs for more info.
    vol : numpy.ndarray
        The image data. Must be NxMxL (or NxMxLxK if each voxel is a tuple).
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename and its contents.
    kwargs : ...
        Further keyword arguments are passed to the writer. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    # Test image
    if isinstance(im, np.ndarray):
        if im.ndim == 3:
            pass
        elif im.ndim == 4 and im.shape[3] < 32:  # How large can a tuple be?
            pass
        else:
            raise ValueError('Image must be 3D, or 4D if each voxel is '
                             'a tuple.')
    else:
        raise ValueError('Image must be a numpy array.')
    
    # Get writer and write first
    writer = get_writer(uri, format, 'v', **kwargs)
    with writer:
        writer.append_data(im)
    
    # Return a result if there is any
    return writer.request.get_result()


## Multiple volumes

def mvolread(uri, format=None, memtest=True, **kwargs):
    """ mvolread(uri, format=None, memtest=True, **kwargs)
    
    Reads multiple volumes from the specified file. Returns a list of
    numpy arrays, each with a dict of meta data at its 'meta' attribute.
    
    Parameters
    ----------
    uri : {str, bytes, file}
        The resource to load the volumes from, e.g. a filename, http address or
        file object, see the docs for more info.
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename and its contents.
    memtest : bool
        If True (default), this function will raise an error if the
        resulting list of images consumes over 1 GB of memory. This
        is to protect the system using so much memory that it needs to
        resort to swapping, and thereby stall the computer. E.g.
        ``mimread('hunger_games.avi')``.
    kwargs : ...
        Further keyword arguments are passed to the reader. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    # Get reader and read all
    reader = read(uri, format, 'V', **kwargs)
    
    ims = []
    nbytes = 0
    for im in reader:
        ims.append(im)
        # Memory check
        nbytes += im.nbytes
        if memtest and nbytes > 1024 * 1024 * 1024:  # pragma: no cover
            ims[:] = []  # clear to free the memory
            raise RuntimeError('imageio.mvolread() has read over 1 GiB of '
                               'image data.\nStopped to avoid memory problems.'
                               ' Use imageio.get_reader() or memtest=False.')
    
    return ims


def mvolwrite(uri, ims, format=None, **kwargs):
    """ mvolwrite(uri, vols, format=None, **kwargs)
    
    Write multiple volumes to the specified file.
    
    Parameters
    ----------
    uri : {str, file}
        The resource to write the volumes to, e.g. a filename or file object,
        see the docs for more info.
    ims : sequence of numpy arrays
        The image data. Each array must be NxMxL (or NxMxLxK if each
        voxel is a tuple).
    format : str
        The format to use to read the file. By default imageio selects
        the appropriate for you based on the filename and its contents.
    kwargs : ...
        Further keyword arguments are passed to the writer. See :func:`.help`
        to see what arguments are available for a particular format.
    """ 
    
    # Get writer
    writer = get_writer(uri, format, 'V', **kwargs)
    with writer:
        
        # Iterate over images (ims may be a generator)
        for im in ims:
            
            # Test image
            if isinstance(im, np.ndarray):
                if im.ndim == 3:
                    pass
                elif im.ndim == 4 and im.shape[3] < 32:
                    pass  # How large can a tuple be?
                else:
                    raise ValueError('Image must be 3D, or 4D if each voxel is'
                                     'a tuple.')
            else:
                raise ValueError('Image must be a numpy array.')
            
            # Add image
            writer.append_data(im)
    
    # Return a result if there is any
    return writer.request.get_result()


## Aliases

read = get_reader
save = get_writer
imsave = imwrite
mimsave = mimwrite
volsave = volwrite
mvolsave = mvolwrite



# -*- coding: utf-8 -*-
# Copyright (c) 2015, imageio contributors
# Copyright (C) 2013, Zach Pincus, Almar Klein and others

""" This module contains generic code to find and load a dynamic library.
"""

from __future__ import absolute_import, print_function, division

import os
import sys
import ctypes


LOCALDIR = os.path.abspath(os.path.dirname(__file__))

# Flag that can be patched / set to True to disable loading non-system libs
SYSTEM_LIBS_ONLY = False


def looks_lib(fname):
    """ Returns True if the given filename looks like a dynamic library.
    Based on extension, but cross-platform and more flexible. 
    """
    fname = fname.lower()
    if sys.platform.startswith('win'):
        return fname.endswith('.dll')
    elif sys.platform.startswith('darwin'):
        return fname.endswith('.dylib')
    else:
        return fname.endswith('.so') or '.so.' in fname


def generate_candidate_libs(lib_names, lib_dirs=None):
    """ Generate a list of candidate filenames of what might be the dynamic
    library corresponding with the given list of names.
    Returns (lib_dirs, lib_paths)
    """
    lib_dirs = lib_dirs or []
    
    # Get system dirs to search
    sys_lib_dirs = ['/lib', 
                    '/usr/lib', 
                    '/usr/lib/x86_64-linux-gnu',
                    '/usr/local/lib', 
                    '/opt/local/lib', ]
    
    # Get Python dirs to search (shared if for Pyzo)
    py_sub_dirs = ['lib', 'DLLs', 'Library/bin', 'shared']    
    py_lib_dirs = [os.path.join(sys.prefix, d) for d in py_sub_dirs]
    if hasattr(sys, 'base_prefix'):
        py_lib_dirs += [os.path.join(sys.base_prefix, d) for d in py_sub_dirs]
    
    # Get user dirs to search (i.e. HOME)
    home_dir = os.path.expanduser('~')
    user_lib_dirs = [os.path.join(home_dir, d) for d in ['lib']]
    
    # Select only the dirs for which a directory exists, and remove duplicates
    potential_lib_dirs = lib_dirs + sys_lib_dirs + py_lib_dirs + user_lib_dirs
    lib_dirs = []
    for ld in potential_lib_dirs:
        if os.path.isdir(ld) and ld not in lib_dirs:
            lib_dirs.append(ld)
    
    # Now attempt to find libraries of that name in the given directory
    # (case-insensitive)
    lib_paths = []
    for lib_dir in lib_dirs:
        # Get files, prefer short names, last version
        files = os.listdir(lib_dir)
        files = reversed(sorted(files))
        files = sorted(files, key=len)
        for lib_name in lib_names:
            # Test all filenames for name and ext
            for fname in files:
                if fname.lower().startswith(lib_name) and looks_lib(fname):
                    lib_paths.append(os.path.join(lib_dir, fname))
    
    # Return (only the items which are files)
    lib_paths = [lp for lp in lib_paths if os.path.isfile(lp)]
    return lib_dirs, lib_paths


def load_lib(exact_lib_names, lib_names, lib_dirs=None):
    """ load_lib(exact_lib_names, lib_names, lib_dirs=None)
    
    Load a dynamic library. 
    
    This function first tries to load the library from the given exact
    names. When that fails, it tries to find the library in common
    locations. It searches for files that start with one of the names
    given in lib_names (case insensitive). The search is performed in
    the given lib_dirs and a set of common library dirs.
    
    Returns ``(ctypes_library, library_path)``
    """
    
    # Checks
    assert isinstance(exact_lib_names, list)
    assert isinstance(lib_names, list)
    if lib_dirs is not None:
        assert isinstance(lib_dirs, list)
    exact_lib_names = [n for n in exact_lib_names if n]
    lib_names = [n for n in lib_names if n]
    
    # Get reference name (for better messages)
    if lib_names:
        the_lib_name = lib_names[0]
    elif exact_lib_names:
        the_lib_name = exact_lib_names[0]
    else:
        raise ValueError("No library name given.")
    
    # Collect filenames of potential libraries
    # First try a few bare library names that ctypes might be able to find
    # in the default locations for each platform. 
    if SYSTEM_LIBS_ONLY:
        lib_dirs, lib_paths = [], []
    else:
        lib_dirs, lib_paths = generate_candidate_libs(lib_names, lib_dirs)
    lib_paths = exact_lib_names + lib_paths
    
    # Select loader 
    if sys.platform.startswith('win'):
        loader = ctypes.windll
    else:
        loader = ctypes.cdll
    
    # Try to load until success
    the_lib = None
    errors = []
    for fname in lib_paths:
        try:
            the_lib = loader.LoadLibrary(fname)
            break
        except Exception:
            # Don't record errors when it couldn't load the library from an
            # exact name -- this fails often, and doesn't provide any useful
            # debugging information anyway, beyond "couldn't find library..."
            if fname not in exact_lib_names:
                # Get exception instance in Python 2.x/3.x compatible manner
                e_type, e_value, e_tb = sys.exc_info()
                del e_tb
                errors.append((fname, e_value))
    
    # No success ...
    if the_lib is None:
        if errors:
            # No library loaded, and load-errors reported for some
            # candidate libs
            err_txt = ['%s:\n%s' % (l, str(e)) for l, e in errors]
            msg = ('One or more %s libraries were found, but ' + 
                   'could not be loaded due to the following errors:\n%s')
            raise OSError(msg % (the_lib_name, '\n\n'.join(err_txt)))
        else:
            # No errors, because no potential libraries found at all!
            msg = 'Could not find a %s library in any of:\n%s'
            raise OSError(msg % (the_lib_name, '\n'.join(lib_dirs)))
    
    # Done
    return the_lib, fname



# -*- coding: utf-8 -*-
# Copyright (c) 2015, imageio contributors

# styletest: ignore E122 E123 E501

"""
Module that contain info about the Pillow formats. The first part of
this module generates this info and writes it to its own bottom half
if run as a script.
"""


def generate_info():  # pragma: no cover
    from urllib.request import urlopen
    import PIL
    from PIL import Image
    
    Image.init()
    
    ids = []
    formats = []
    docs = {}
    
    # Collect formats and their summary from plugin modules
    for mod_name in dir(PIL):
        if 'ImagePlugin' in mod_name:
            mod = getattr(PIL, mod_name)
            for ob_name in dir(mod):
                ob = getattr(mod, ob_name)
                if isinstance(ob, type) and issubclass(ob, Image.Image):
                    if ob.format in ids:
                        print('Found duplicate for', ob.format)
                    else:
                        ids.append(ob.format)
                        formats.append((ob.format, ob.format_description))
    
    # Add extension info
    for i in range(len(formats)):
        id, summary = formats[i]
        ext = ' '.join([e for e in Image.EXTENSION if Image.EXTENSION[e] == id])
        formats[i] = id, summary, ext
    
    # Get documentation of formats
    url = 'https://raw.githubusercontent.com/python-pillow/Pillow/master/docs/handbook/image-file-formats.rst'  # noqa
    lines = urlopen(url).read().decode().splitlines()
    lines.append('End')
    lines.append('---')  # for the end
    
    # Parse documentation
    cur_name = ''
    cur_part = []
    for i in range(len(lines)):
        line = lines[i]
        if line.startswith(('^^^', '---', '===')):
            if cur_name and cur_name in ids:
                text = '\n'.join(cur_part[:-1])
                text = text.replace('versionadded::', 'versionadded:: Pillow ')
                text = text.replace('Image.open`', 'Image.write`')
                docs[cur_name] = text
            cur_part = []
            cur_name = lines[i-1].strip().replace(' ', '').upper()
        else:
            cur_part.append('    ' + line)
    
    # Fill in the blancs
    for id in ids:
        if id in docs:
            docs[id] = '*This is a copy from the Pillow docs.*\n\n' + docs[id]
        else:
            docs[id] = 'No docs for %s.' % id
            print('no docs for', id)
    
    # Sort before writing
    formats.sort(key=lambda x: x[0])
    ids.sort()
    
    # Read file ...
    code = open(__file__, 'rb').read().decode()
    code, divider, _ = code.partition('## BELOW IS ' + 'AUTOGENERATED')
    code += divider + '\n\n'
    
    # Write formats
    code += 'pillow_formats = [\n'
    for i in range(len(formats)):
        print(formats[i])
        code += '    (%r, %r, %r),\n' % formats[i]
    code += '    ]\n\n\n'
    
    # Write docs
    code += 'pillow_docs = {\n'
    for id in ids:
        code += '%r:\nu"""%s""",\n' % (id, docs[id])
    code += '}\n'
    
    # Write back
    with open(__file__, 'wb') as f:
        f.write(code.encode())


if __name__ == '__main__':
    generate_info()


## BELOW IS AUTOGENERATED

pillow_formats = [
    ('BMP', 'Windows Bitmap', '.bmp'),
    ('BUFR', 'BUFR', '.bufr'),
    ('CUR', 'Windows Cursor', '.cur'),
    ('DCX', 'Intel DCX', '.dcx'),
    ('DDS', 'DirectDraw Surface', '.dds'),
    ('DIB', 'Windows Bitmap', ''),
    ('EPS', 'Encapsulated Postscript', '.ps .eps'),
    ('FITS', 'FITS', '.fit .fits'),
    ('FLI', 'Autodesk FLI/FLC Animation', '.fli .flc'),
    ('FPX', 'FlashPix', '.fpx'),
    ('FTEX', 'Texture File Format (IW2:EOC)', '.ftc .ftu'),
    ('GBR', 'GIMP brush file', '.gbr'),
    ('GIF', 'Compuserve GIF', '.gif'),
    ('GRIB', 'GRIB', '.grib'),
    ('HDF5', 'HDF5', '.h5 .hdf'),
    ('ICNS', 'Mac OS icns resource', '.icns'),
    ('ICO', 'Windows Icon', '.ico'),
    ('IM', 'IFUNC Image Memory', '.im'),
    ('IMT', 'IM Tools', ''),
    ('IPTC', 'IPTC/NAA', '.iim'),
    ('JPEG', 'JPEG (ISO 10918)', '.jfif .jpe .jpg .jpeg'),
    ('JPEG2000', 'JPEG 2000 (ISO 15444)', '.jp2 .j2k .jpc .jpf .jpx .j2c'),
    ('MCIDAS', 'McIdas area file', ''),
    ('MIC', 'Microsoft Image Composer', '.mic'),
    ('MPEG', 'MPEG', '.mpg .mpeg'),
    ('MPO', 'MPO (CIPA DC-007)', '.mpo'),
    ('MSP', 'Windows Paint', '.msp'),
    ('PCD', 'Kodak PhotoCD', '.pcd'),
    ('PCX', 'Paintbrush', '.pcx'),
    ('PIXAR', 'PIXAR raster image', '.pxr'),
    ('PNG', 'Portable network graphics', '.png'),
    ('PPM', 'Pbmplus image', '.pbm .pgm .ppm'),
    ('PSD', 'Adobe Photoshop', '.psd'),
    ('SGI', 'SGI Image File Format', '.bw .rgb .rgba .sgi'),
    ('SPIDER', 'Spider 2D image', ''),
    ('SUN', 'Sun Raster File', '.ras'),
    ('TGA', 'Targa', '.tga'),
    ('TIFF', 'Adobe TIFF', '.tif .tiff'),
    ('WMF', 'Windows Metafile', '.wmf .emf'),
    ('XBM', 'X11 Bitmap', '.xbm'),
    ('XPM', 'X11 Pixel Map', '.xpm'),
    ('XVThumb', 'XV thumbnail image', ''),
    ]


pillow_docs = {
'BMP':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads and writes Windows and OS/2 BMP files containing ``1``, ``L``, ``P``,
    or ``RGB`` data. 16-colour images are read as ``P`` images. Run-length encoding
    is not supported.
    
    The :py:meth:`~PIL.Image.Image.write` method sets the following
    :py:attr:`~PIL.Image.Image.info` properties:
    
    **compression**
        Set to ``bmp_rle`` if the file is run-length encoded.
    """,
'BUFR':
u"""*This is a copy from the Pillow docs.*

    
    .. versionadded:: Pillow  1.1.3
    
    PIL provides a stub driver for BUFR files.
    
    To add read or write support to your application, use
    :py:func:`PIL.BufrStubImagePlugin.register_handler`.
    """,
'CUR':
u"""*This is a copy from the Pillow docs.*

    
    CUR is used to store cursors on Windows. The CUR decoder reads the largest
    available cursor. Animated cursors are not supported.
    """,
'DCX':
u"""*This is a copy from the Pillow docs.*

    
    DCX is a container file format for PCX files, defined by Intel. The DCX format
    is commonly used in fax applications. The DCX decoder can read files containing
    ``1``, ``L``, ``P``, or ``RGB`` data.
    
    When the file is opened, only the first image is read. You can use
    :py:meth:`~file.seek` or :py:mod:`~PIL.ImageSequence` to read other images.
    
    """,
'DDS':
u"""*This is a copy from the Pillow docs.*

    
    DDS is a popular container texture format used in video games and natively
    supported by DirectX.
    Currently, DXT1, DXT3, and DXT5 pixel formats are supported and only in ``RGBA``
    mode.
    
    .. versionadded:: Pillow  3.4.0 DXT3
    """,
'DIB':
u"""No docs for DIB.""",
'EPS':
u"""*This is a copy from the Pillow docs.*

    
    PIL identifies EPS files containing image data, and can read files that contain
    embedded raster images (ImageData descriptors). If Ghostscript is available,
    other EPS files can be read as well. The EPS driver can also write EPS
    images. The EPS driver can read EPS images in ``L``, ``LAB``, ``RGB`` and
    ``CMYK`` mode, but Ghostscript may convert the images to ``RGB`` mode rather
    than leaving them in the original color space. The EPS driver can write images
    in ``L``, ``RGB`` and ``CMYK`` modes.
    
    If Ghostscript is available, you can call the :py:meth:`~PIL.Image.Image.load`
    method with the following parameter to affect how Ghostscript renders the EPS
    
    **scale**
        Affects the scale of the resultant rasterized image. If the EPS suggests
        that the image be rendered at 100px x 100px, setting this parameter to
        2 will make the Ghostscript render a 200px x 200px image instead. The
        relative position of the bounding box is maintained::
    
            im = Image.open(...)
            im.size #(100,100)
            im.load(scale=2)
            im.size #(200,200)
    """,
'FITS':
u"""*This is a copy from the Pillow docs.*

    
    .. versionadded:: Pillow  1.1.5
    
    PIL provides a stub driver for FITS files.
    
    To add read or write support to your application, use
    :py:func:`PIL.FitsStubImagePlugin.register_handler`.
    """,
'FLI':
u"""No docs for FLI.""",
'FPX':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads Kodak FlashPix files. In the current version, only the highest
    resolution image is read from the file, and the viewing transform is not taken
    into account.
    
    .. note::
    
        To enable full FlashPix support, you need to build and install the IJG JPEG
        library before building the Python Imaging Library. See the distribution
        README for details.
    """,
'FTEX':
u"""*This is a copy from the Pillow docs.*

    
    .. versionadded:: Pillow  3.2.0
    
    The FTEX decoder reads textures used for 3D objects in
    Independence War 2: Edge Of Chaos. The plugin reads a single texture
    per file, in the compressed and uncompressed formats.
    """,
'GBR':
u"""*This is a copy from the Pillow docs.*

    
    The GBR decoder reads GIMP brush files, version 1 and 2.
    
    The :py:meth:`~PIL.Image.Image.write` method sets the following
    :py:attr:`~PIL.Image.Image.info` properties:
    
    **comment**
        The brush name.
    
    **spacing**
        The spacing between the brushes, in pixels. Version 2 only.
    
    GD
    ^^
    
    PIL reads uncompressed GD files. Note that this file format cannot be
    automatically identified, so you must use :py:func:`PIL.GdImageFile.open` to
    read such a file.
    
    The :py:meth:`~PIL.Image.Image.write` method sets the following
    :py:attr:`~PIL.Image.Image.info` properties:
    
    **transparency**
        Transparency color index. This key is omitted if the image is not
        transparent.
    """,
'GIF':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads GIF87a and GIF89a versions of the GIF file format. The library writes
    run-length encoded files in GIF87a by default, unless GIF89a features
    are used or GIF89a is already in use.
    
    Note that GIF files are always read as grayscale (``L``)
    or palette mode (``P``) images.
    
    The :py:meth:`~PIL.Image.Image.write` method sets the following
    :py:attr:`~PIL.Image.Image.info` properties:
    
    **background**
        Default background color (a palette color index).
    
    **transparency**
        Transparency color index. This key is omitted if the image is not
        transparent.
    
    **version**
        Version (either ``GIF87a`` or ``GIF89a``).
    
    **duration**
        May not be present. The time to display the current frame
        of the GIF, in milliseconds.
    
    **loop**
        May not be present. The number of times the GIF should loop.
    
    Reading sequences
    ~~~~~~~~~~~~~~~~~
    
    The GIF loader supports the :py:meth:`~file.seek` and :py:meth:`~file.tell`
    methods. You can seek to the next frame (``im.seek(im.tell() + 1)``), or rewind
    the file by seeking to the first frame. Random access is not supported.
    
    ``im.seek()`` raises an ``EOFError`` if you try to seek after the last frame.
    
    Saving
    ~~~~~~
    
    When calling :py:meth:`~PIL.Image.Image.save`, the following options
    are available::
    
        im.save(out, save_all=True, append_images=[im1, im2, ...])
    
    **save_all**
        If present and true, all frames of the image will be saved. If
        not, then only the first frame of a multiframe image will be saved.
    
    **append_images**
        A list of images to append as additional frames. Each of the
        images in the list can be single or multiframe images.
    
    **duration**
        The display duration of each frame of the multiframe gif, in
        milliseconds. Pass a single integer for a constant duration, or a
        list or tuple to set the duration for each frame separately.
    
    **loop**
        Integer number of times the GIF should loop.
    
    **optimize**
        If present and true, attempt to compress the palette by
        eliminating unused colors. This is only useful if the palette can
        be compressed to the next smaller power of 2 elements.
    
    **palette** 
        Use the specified palette for the saved image. The palette should
        be a bytes or bytearray object containing the palette entries in
        RGBRGB... form. It should be no more than 768 bytes. Alternately,
        the palette can be passed in as an
        :py:class:`PIL.ImagePalette.ImagePalette` object.
    
    Reading local images
    ~~~~~~~~~~~~~~~~~~~~
    
    The GIF loader creates an image memory the same size as the GIF file’s *logical
    screen size*, and pastes the actual pixel data (the *local image*) into this
    image. If you only want the actual pixel rectangle, you can manipulate the
    :py:attr:`~PIL.Image.Image.size` and :py:attr:`~PIL.Image.Image.tile`
    attributes before loading the file::
    
        im = Image.open(...)
    
        if im.tile[0][0] == "gif":
            # only read the first "local image" from this GIF file
            tag, (x0, y0, x1, y1), offset, extra = im.tile[0]
            im.size = (x1 - x0, y1 - y0)
            im.tile = [(tag, (0, 0) + im.size, offset, extra)]
    """,
'GRIB':
u"""*This is a copy from the Pillow docs.*

    
    .. versionadded:: Pillow  1.1.5
    
    PIL provides a stub driver for GRIB files.
    
    The driver requires the file to start with a GRIB header. If you have files
    with embedded GRIB data, or files with multiple GRIB fields, your application
    has to seek to the header before passing the file handle to PIL.
    
    To add read or write support to your application, use
    :py:func:`PIL.GribStubImagePlugin.register_handler`.
    """,
'HDF5':
u"""*This is a copy from the Pillow docs.*

    
    .. versionadded:: Pillow  1.1.5
    
    PIL provides a stub driver for HDF5 files.
    
    To add read or write support to your application, use
    :py:func:`PIL.Hdf5StubImagePlugin.register_handler`.
    """,
'ICNS':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads and (macOS only) writes macOS ``.icns`` files.  By default, the
    largest available icon is read, though you can override this by setting the
    :py:attr:`~PIL.Image.Image.size` property before calling
    :py:meth:`~PIL.Image.Image.load`.  The :py:meth:`~PIL.Image.Image.write` method
    sets the following :py:attr:`~PIL.Image.Image.info` property:
    
    **sizes**
        A list of supported sizes found in this icon file; these are a
        3-tuple, ``(width, height, scale)``, where ``scale`` is 2 for a retina
        icon and 1 for a standard icon.  You *are* permitted to use this 3-tuple
        format for the :py:attr:`~PIL.Image.Image.size` property if you set it
        before calling :py:meth:`~PIL.Image.Image.load`; after loading, the size
        will be reset to a 2-tuple containing pixel dimensions (so, e.g. if you
        ask for ``(512, 512, 2)``, the final value of
        :py:attr:`~PIL.Image.Image.size` will be ``(1024, 1024)``).
    """,
'ICO':
u"""*This is a copy from the Pillow docs.*

    
    ICO is used to store icons on Windows. The largest available icon is read.
    
    The :py:meth:`~PIL.Image.Image.save` method supports the following options:
    
    **sizes**
        A list of sizes including in this ico file; these are a 2-tuple,
        ``(width, height)``; Default to ``[(16, 16), (24, 24), (32, 32), (48, 48),
        (64, 64), (128, 128), (255, 255)]``. Any sizes bigger than the original
        size or 255 will be ignored.
    
    IM
    ^^
    
    IM is a format used by LabEye and other applications based on the IFUNC image
    processing library. The library reads and writes most uncompressed interchange
    versions of this format.
    
    IM is the only format that can store all internal PIL formats.
    """,
'IM':
u"""No docs for IM.""",
'IMT':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads Image Tools images containing ``L`` data.
    """,
'IPTC':
u"""No docs for IPTC.""",
'JPEG':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads JPEG, JFIF, and Adobe JPEG files containing ``L``, ``RGB``, or
    ``CMYK`` data. It writes standard and progressive JFIF files.
    
    Using the :py:meth:`~PIL.Image.Image.draft` method, you can speed things up by
    converting ``RGB`` images to ``L``, and resize images to 1/2, 1/4 or 1/8 of
    their original size while loading them.
    
    The :py:meth:`~PIL.Image.Image.write` method may set the following
    :py:attr:`~PIL.Image.Image.info` properties if available:
    
    **jfif**
        JFIF application marker found. If the file is not a JFIF file, this key is
        not present.
    
    **jfif_version**
        A tuple representing the jfif version, (major version, minor version).
    
    **jfif_density**
        A tuple representing the pixel density of the image, in units specified
        by jfif_unit.
    
    **jfif_unit**
        Units for the jfif_density:
    
        * 0 - No Units
        * 1 - Pixels per Inch
        * 2 - Pixels per Centimeter
    
    **dpi**
        A tuple representing the reported pixel density in pixels per inch, if
        the file is a jfif file and the units are in inches.
    
    **adobe**
        Adobe application marker found. If the file is not an Adobe JPEG file, this
        key is not present.
    
    **adobe_transform**
        Vendor Specific Tag.
    
    **progression**
        Indicates that this is a progressive JPEG file.
    
    **icc_profile**
        The ICC color profile for the image.
    
    **exif**
        Raw EXIF data from the image.
    
    
    The :py:meth:`~PIL.Image.Image.save` method supports the following options:
    
    **quality**
        The image quality, on a scale from 1 (worst) to 95 (best). The default is
        75. Values above 95 should be avoided; 100 disables portions of the JPEG
        compression algorithm, and results in large files with hardly any gain in
        image quality.
    
    **optimize**
        If present and true, indicates that the encoder should make an extra pass
        over the image in order to select optimal encoder settings.
    
    **progressive**
        If present and true, indicates that this image should be stored as a
        progressive JPEG file.
    
    **dpi**
        A tuple of integers representing the pixel density, ``(x,y)``.
    
    **icc_profile**
        If present and true, the image is stored with the provided ICC profile.
        If this parameter is not provided, the image will be saved with no profile
        attached. To preserve the existing profile::
    
            im.save(filename, 'jpeg', icc_profile=im.info.get('icc_profile'))
    
    **exif**
        If present, the image will be stored with the provided raw EXIF data.
    
    **subsampling**
        If present, sets the subsampling for the encoder.
    
        * ``keep``: Only valid for JPEG files, will retain the original image setting.
        * ``4:4:4``, ``4:2:2``, ``4:1:1``: Specific sampling values
        * ``-1``: equivalent to ``keep``
        * ``0``: equivalent to ``4:4:4``
        * ``1``: equivalent to ``4:2:2``
        * ``2``: equivalent to ``4:1:1``
    
    **qtables**
        If present, sets the qtables for the encoder. This is listed as an
        advanced option for wizards in the JPEG documentation. Use with
        caution. ``qtables`` can be one of several types of values:
    
        *  a string, naming a preset, e.g. ``keep``, ``web_low``, or ``web_high``
        *  a list, tuple, or dictionary (with integer keys =
           range(len(keys))) of lists of 64 integers. There must be
           between 2 and 4 tables.
    
        .. versionadded:: Pillow  2.5.0
    
    
    .. note::
    
        To enable JPEG support, you need to build and install the IJG JPEG library
        before building the Python Imaging Library. See the distribution README for
        details.
    """,
'JPEG2000':
u"""*This is a copy from the Pillow docs.*

    
    .. versionadded:: Pillow  2.4.0
    
    PIL reads and writes JPEG 2000 files containing ``L``, ``LA``, ``RGB`` or
    ``RGBA`` data.  It can also read files containing ``YCbCr`` data, which it
    converts on read into ``RGB`` or ``RGBA`` depending on whether or not there is
    an alpha channel.  PIL supports JPEG 2000 raw codestreams (``.j2k`` files), as
    well as boxed JPEG 2000 files (``.j2p`` or ``.jpx`` files).  PIL does *not*
    support files whose components have different sampling frequencies.
    
    When loading, if you set the ``mode`` on the image prior to the
    :py:meth:`~PIL.Image.Image.load` method being invoked, you can ask PIL to
    convert the image to either ``RGB`` or ``RGBA`` rather than choosing for
    itself.  It is also possible to set ``reduce`` to the number of resolutions to
    discard (each one reduces the size of the resulting image by a factor of 2),
    and ``layers`` to specify the number of quality layers to load.
    
    The :py:meth:`~PIL.Image.Image.save` method supports the following options:
    
    **offset**
        The image offset, as a tuple of integers, e.g. (16, 16)
    
    **tile_offset**
        The tile offset, again as a 2-tuple of integers.
    
    **tile_size**
        The tile size as a 2-tuple.  If not specified, or if set to None, the
        image will be saved without tiling.
    
    **quality_mode**
        Either `"rates"` or `"dB"` depending on the units you want to use to
        specify image quality.
    
    **quality_layers**
        A sequence of numbers, each of which represents either an approximate size
        reduction (if quality mode is `"rates"`) or a signal to noise ratio value
        in decibels.  If not specified, defaults to a single layer of full quality.
    
    **num_resolutions**
        The number of different image resolutions to be stored (which corresponds
        to the number of Discrete Wavelet Transform decompositions plus one).
    
    **codeblock_size**
        The code-block size as a 2-tuple.  Minimum size is 4 x 4, maximum is 1024 x
        1024, with the additional restriction that no code-block may have more
        than 4096 coefficients (i.e. the product of the two numbers must be no
        greater than 4096).
    
    **precinct_size**
        The precinct size as a 2-tuple.  Must be a power of two along both axes,
        and must be greater than the code-block size.
    
    **irreversible**
        If ``True``, use the lossy Irreversible Color Transformation
        followed by DWT 9-7.  Defaults to ``False``, which means to use the
        Reversible Color Transformation with DWT 5-3.
    
    **progression**
        Controls the progression order; must be one of ``"LRCP"``, ``"RLCP"``,
        ``"RPCL"``, ``"PCRL"``, ``"CPRL"``.  The letters stand for Component,
        Position, Resolution and Layer respectively and control the order of
        encoding, the idea being that e.g. an image encoded using LRCP mode can
        have its quality layers decoded as they arrive at the decoder, while one
        encoded using RLCP mode will have increasing resolutions decoded as they
        arrive, and so on.
    
    **cinema_mode**
        Set the encoder to produce output compliant with the digital cinema
        specifications.  The options here are ``"no"`` (the default),
        ``"cinema2k-24"`` for 24fps 2K, ``"cinema2k-48"`` for 48fps 2K, and
        ``"cinema4k-24"`` for 24fps 4K.  Note that for compliant 2K files,
        *at least one* of your image dimensions must match 2048 x 1080, while
        for compliant 4K files, *at least one* of the dimensions must match
        4096 x 2160.
    
    .. note::
    
       To enable JPEG 2000 support, you need to build and install the OpenJPEG
       library, version 2.0.0 or higher, before building the Python Imaging
       Library.
    
       Windows users can install the OpenJPEG binaries available on the
       OpenJPEG website, but must add them to their PATH in order to use PIL (if
       you fail to do this, you will get errors about not being able to load the
       ``_imaging`` DLL).
    """,
'MCIDAS':
u"""*This is a copy from the Pillow docs.*

    
    PIL identifies and reads 8-bit McIdas area files.
    """,
'MIC':
u"""*This is a copy from the Pillow docs.*

    
    PIL identifies and reads Microsoft Image Composer (MIC) files. When opened, the
    first sprite in the file is loaded. You can use :py:meth:`~file.seek` and
    :py:meth:`~file.tell` to read other sprites from the file.
    """,
'MPEG':
u"""*This is a copy from the Pillow docs.*

    
    PIL identifies MPEG files.
    """,
'MPO':
u"""*This is a copy from the Pillow docs.*

    
    Pillow identifies and reads Multi Picture Object (MPO) files, loading the primary
    image when first opened. The :py:meth:`~file.seek` and :py:meth:`~file.tell`
    methods may be used to read other pictures from the file. The pictures are
    zero-indexed and random access is supported.
    """,
'MSP':
u"""*This is a copy from the Pillow docs.*

    
    PIL identifies and reads MSP files from Windows 1 and 2. The library writes
    uncompressed (Windows 1) versions of this format.
    """,
'PCD':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads PhotoCD files containing ``RGB`` data. This only reads the 768x512
    resolution image from the file. Higher resolutions are encoded in a proprietary
    encoding.
    """,
'PCX':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads and writes PCX files containing ``1``, ``L``, ``P``, or ``RGB`` data.
    """,
'PIXAR':
u"""*This is a copy from the Pillow docs.*

    
    PIL provides limited support for PIXAR raster files. The library can identify
    and read “dumped” RGB files.
    
    The format code is ``PIXAR``.
    """,
'PNG':
u"""*This is a copy from the Pillow docs.*

    
    PIL identifies, reads, and writes PNG files containing ``1``, ``L``, ``P``,
    ``RGB``, or ``RGBA`` data. Interlaced files are supported as of v1.1.7.
    
    The :py:meth:`~PIL.Image.Image.write` method sets the following
    :py:attr:`~PIL.Image.Image.info` properties, when appropriate:
    
    **gamma**
        Gamma, given as a floating point number.
    
    **transparency**
        For ``P`` images: Either the palette index for full transparent pixels,
        or a byte string with alpha values for each palette entry.
    
        For ``L`` and ``RGB`` images, the color that represents full transparent
        pixels in this image.
    
        This key is omitted if the image is not a transparent palette image.
    
    ``Open`` also sets ``Image.text`` to a list of the values of the
    ``tEXt``, ``zTXt``, and ``iTXt`` chunks of the PNG image. Individual
    compressed chunks are limited to a decompressed size of
    ``PngImagePlugin.MAX_TEXT_CHUNK``, by default 1MB, to prevent
    decompression bombs. Additionally, the total size of all of the text
    chunks is limited to ``PngImagePlugin.MAX_TEXT_MEMORY``, defaulting to
    64MB.
    
    The :py:meth:`~PIL.Image.Image.save` method supports the following options:
    
    **optimize**
        If present and true, instructs the PNG writer to make the output file as
        small as possible. This includes extra processing in order to find optimal
        encoder settings.
    
    **transparency**
        For ``P``, ``L``, and ``RGB`` images, this option controls what
        color image to mark as transparent.
    
        For ``P`` images, this can be a either the palette index,
        or a byte string with alpha values for each palette entry.
    
    **dpi**
        A tuple of two numbers corresponding to the desired dpi in each direction.
    
    **pnginfo**
        A :py:class:`PIL.PngImagePlugin.PngInfo` instance containing text tags.
    
    **compress_level**
        ZLIB compression level, a number between 0 and 9: 1 gives best speed,
        9 gives best compression, 0 gives no compression at all. Default is 6.
        When ``optimize`` option is True ``compress_level`` has no effect
        (it is set to 9 regardless of a value passed).
    
    **icc_profile**
        The ICC Profile to include in the saved file.
    
    **bits (experimental)**
        For ``P`` images, this option controls how many bits to store. If omitted,
        the PNG writer uses 8 bits (256 colors).
    
    **dictionary (experimental)**
        Set the ZLIB encoder dictionary.
    
    .. note::
    
        To enable PNG support, you need to build and install the ZLIB compression
        library before building the Python Imaging Library. See the installation
        documentation for details.
    """,
'PPM':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads and writes PBM, PGM and PPM files containing ``1``, ``L`` or ``RGB``
    data.
    """,
'PSD':
u"""*This is a copy from the Pillow docs.*

    
    PIL identifies and reads PSD files written by Adobe Photoshop 2.5 and 3.0.
    
    """,
'SGI':
u"""*This is a copy from the Pillow docs.*

    
    Pillow reads and writes uncompressed ``L``, ``RGB``, and ``RGBA`` files.
    
    """,
'SPIDER':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads and writes SPIDER image files of 32-bit floating point data
    ("F;32F").
    
    PIL also reads SPIDER stack files containing sequences of SPIDER images. The
    :py:meth:`~file.seek` and :py:meth:`~file.tell` methods are supported, and
    random access is allowed.
    
    The :py:meth:`~PIL.Image.Image.write` method sets the following attributes:
    
    **format**
        Set to ``SPIDER``
    
    **istack**
        Set to 1 if the file is an image stack, else 0.
    
    **nimages**
        Set to the number of images in the stack.
    
    A convenience method, :py:meth:`~PIL.Image.Image.convert2byte`, is provided for
    converting floating point data to byte data (mode ``L``)::
    
        im = Image.open('image001.spi').convert2byte()
    
    Writing files in SPIDER format
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    The extension of SPIDER files may be any 3 alphanumeric characters. Therefore
    the output format must be specified explicitly::
    
        im.save('newimage.spi', format='SPIDER')
    
    For more information about the SPIDER image processing package, see the
    `SPIDER homepage`_ at `Wadsworth Center`_.
    
    .. _SPIDER homepage: https://spider.wadsworth.org/spider_doc/spider/docs/spider.html
    .. _Wadsworth Center: https://www.wadsworth.org/
    """,
'SUN':
u"""No docs for SUN.""",
'TGA':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads 24- and 32-bit uncompressed and run-length encoded TGA files.
    """,
'TIFF':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads and writes TIFF files. It can read both striped and tiled images,
    pixel and plane interleaved multi-band images, and either uncompressed, or
    Packbits, LZW, or JPEG compressed images.
    
    If you have libtiff and its headers installed, PIL can read and write many more
    kinds of compressed TIFF files. If not, PIL will always write uncompressed
    files.
    
    The :py:meth:`~PIL.Image.Image.write` method sets the following
    :py:attr:`~PIL.Image.Image.info` properties:
    
    **compression**
        Compression mode.
    
        .. versionadded:: Pillow  2.0.0
    
    **dpi**
        Image resolution as an ``(xdpi, ydpi)`` tuple, where applicable. You can use
        the :py:attr:`~PIL.Image.Image.tag` attribute to get more detailed
        information about the image resolution.
    
        .. versionadded:: Pillow  1.1.5
    
    **resolution**
        Image resolution as an ``(xres, yres)`` tuple, where applicable. This is a
        measurement in whichever unit is specified by the file.
    
        .. versionadded:: Pillow  1.1.5
    
    
    The :py:attr:`~PIL.Image.Image.tag_v2` attribute contains a dictionary
    of TIFF metadata. The keys are numerical indexes from
    :py:attr:`~PIL.TiffTags.TAGS_V2`.  Values are strings or numbers for single
    items, multiple values are returned in a tuple of values. Rational
    numbers are returned as a :py:class:`~PIL.TiffImagePlugin.IFDRational`
    object.
    
        .. versionadded:: Pillow  3.0.0
    
    For compatibility with legacy code, the
    :py:attr:`~PIL.Image.Image.tag` attribute contains a dictionary of
    decoded TIFF fields as returned prior to version 3.0.0.  Values are
    returned as either strings or tuples of numeric values. Rational
    numbers are returned as a tuple of ``(numerator, denominator)``.
    
        .. deprecated:: 3.0.0
    
    
    Saving Tiff Images
    ~~~~~~~~~~~~~~~~~~
    
    The :py:meth:`~PIL.Image.Image.save` method can take the following keyword arguments:
    
    **save_all**
        If true, Pillow will save all frames of the image to a multiframe tiff document.
    
        .. versionadded:: Pillow  3.4.0
    
    **tiffinfo**
        A :py:class:`~PIL.TiffImagePlugin.ImageFileDirectory_v2` object or dict
        object containing tiff tags and values. The TIFF field type is
        autodetected for Numeric and string values, any other types
        require using an :py:class:`~PIL.TiffImagePlugin.ImageFileDirectory_v2`
        object and setting the type in
        :py:attr:`~PIL.TiffImagePlugin.ImageFileDirectory_v2.tagtype` with
        the appropriate numerical value from
        ``TiffTags.TYPES``.
    
        .. versionadded:: Pillow  2.3.0
    
        Metadata values that are of the rational type should be passed in
        using a :py:class:`~PIL.TiffImagePlugin.IFDRational` object.
    
        .. versionadded:: Pillow  3.1.0
    
        For compatibility with legacy code, a
        :py:class:`~PIL.TiffImagePlugin.ImageFileDirectory_v1` object may
        be passed in this field. However, this is deprecated.
    
        .. versionadded:: Pillow  3.0.0
    
     .. note::
    
        Only some tags are currently supported when writing using
        libtiff. The supported list is found in
        :py:attr:`~PIL:TiffTags.LIBTIFF_CORE`.
    
    **compression**
        A string containing the desired compression method for the
        file. (valid only with libtiff installed) Valid compression
        methods are: ``None``, ``"tiff_ccitt"``, ``"group3"``,
        ``"group4"``, ``"tiff_jpeg"``, ``"tiff_adobe_deflate"``,
        ``"tiff_thunderscan"``, ``"tiff_deflate"``, ``"tiff_sgilog"``,
        ``"tiff_sgilog24"``, ``"tiff_raw_16"``
    
    These arguments to set the tiff header fields are an alternative to
    using the general tags available through tiffinfo.
    
    **description**
    
    **software**
    
    **date_time**
    
    **artist**
    
    **copyright**
        Strings
    
    **resolution_unit**
        A string of "inch", "centimeter" or "cm"
    
    **resolution**
    
    **x_resolution**
    
    **y_resolution**
    
    **dpi**
        Either a Float, 2 tuple of (numerator, denominator) or a
        :py:class:`~PIL.TiffImagePlugin.IFDRational`. Resolution implies
        an equal x and y resolution, dpi also implies a unit of inches.
    
    """,
'WMF':
u"""*This is a copy from the Pillow docs.*

    
    PIL can identify playable WMF files.
    
    In PIL 1.1.4 and earlier, the WMF driver provides some limited rendering
    support, but not enough to be useful for any real application.
    
    In PIL 1.1.5 and later, the WMF driver is a stub driver. To add WMF read or
    write support to your application, use
    :py:func:`PIL.WmfImagePlugin.register_handler` to register a WMF handler.
    
    ::
    
        from PIL import Image
        from PIL import WmfImagePlugin
    
        class WmfHandler:
            def open(self, im):
                ...
            def load(self, im):
                ...
                return image
            def save(self, im, fp, filename):
                ...
    
        wmf_handler = WmfHandler()
    
        WmfImagePlugin.register_handler(wmf_handler)
    
        im = Image.open("sample.wmf")""",
'XBM':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads and writes X bitmap files (mode ``1``).
    """,
'XPM':
u"""*This is a copy from the Pillow docs.*

    
    PIL reads X pixmap files (mode ``P``) with 256 colors or less.
    
    The :py:meth:`~PIL.Image.Image.write` method sets the following
    :py:attr:`~PIL.Image.Image.info` properties:
    
    **transparency**
        Transparency color index. This key is omitted if the image is not
        transparent.
    """,
'XVThumb':
u"""No docs for XVThumb.""",
}



# -*- coding: utf-8 -*-
# Copyright (c) 2016, imageio contributors
# imageio is distributed under the terms of the (new) BSD License.

""" Plugin that wraps the the Pillow library.
"""

from __future__ import absolute_import, print_function, division

from warnings import warn
import numpy as np
import threading

from .. import formats
from ..core import Format, image_as_uint

# Get info about pillow formats without having to import PIL
from .pillow_info import pillow_formats, pillow_docs

# todo: Pillow ImageGrab module supports grabbing the screen on Win and OSX.


class PillowFormat(Format):
    """
    Base format class for Pillow formats.
    """
    
    _pillow_imported = False
    _Image = None
    _modes = 'i'
    _description = ''

    def __init__(self, *args, **kwargs):
        super(PillowFormat, self).__init__(*args, **kwargs)
        # Used to synchronize _init_pillow(), see #244
        self._lock = threading.RLock()

    @property
    def plugin_id(self):
        """ The PIL plugin id.
        """
        return self._plugin_id  # Set when format is created
    
    def _init_pillow(self):
        with self._lock:
            if not self._pillow_imported:
                self._pillow_imported = True  # more like tried to import
                import PIL
                if not hasattr(PIL, 'PILLOW_VERSION'):  # pragma: no cover
                    raise ImportError('Imageio Pillow requires '
                                      'Pillow, not PIL!')
                from PIL import Image
                self._Image = Image
            elif self._Image is None:  # pragma: no cover
                raise RuntimeError('Imageio Pillow plugin requires '
                                   'Pillow lib.')
            Image = self._Image

        if self.plugin_id in ('PNG', 'JPEG', 'BMP', 'GIF', 'PPM'):
            Image.preinit()
        else:
            Image.init()
        return Image
    
    def _can_read(self, request):
        Image = self._init_pillow()
        if request.mode[1] in (self.modes + '?'):
            if self.plugin_id in Image.OPEN:
                factory, accept = Image.OPEN[self.plugin_id]
                if accept:
                    if request.firstbytes and accept(request.firstbytes):
                        return True
    
    def _can_write(self, request):
        Image = self._init_pillow()
        if request.mode[1] in (self.modes + '?'):
            if request.filename.lower().endswith(self.extensions):
                if self.plugin_id in Image.SAVE:
                    return True
    
    class Reader(Format.Reader):
    
        def _open(self, **kwargs):
            Image = self.format._init_pillow()
            try:
                factory, accept = Image.OPEN[self.format.plugin_id]
            except KeyError:
                raise RuntimeError('Format %s cannot read images.' %
                                   self.format.name)
            self._fp = self.request.get_file()
            self._im = factory(self._fp, '')
            if hasattr(Image, '_decompression_bomb_check'):
                Image._decompression_bomb_check(self._im.size)
            pil_try_read(self._im)
            self._grayscale = _palette_is_grayscale(self._im)
            # Set length
            self._length = 1
            if hasattr(self._im, 'n_frames'):
                self._length = self._im.n_frames
        
        def _close(self):
            if hasattr(self._im, 'close'):  # see issue #216
                self._im.close()
            # request object handled closing the _fp
        
        def _get_length(self):
            return self._length
        
        def _seek(self, index):
            try:
                self._im.seek(index)
            except EOFError:
                raise IndexError('Could not seek to index %i' % index)
        
        def _get_data(self, index):
            if index >= self._length:
                raise IndexError('Image index %i > %i' % (index, self._length))
            i = self._im.tell()
            if i > index:
                self._seek(index)  # just try
            else:
                while i < index:  # some formats need to be read in sequence
                    i += 1
                    self._seek(i)
            self._im.getdata()[0]
            im = pil_get_frame(self._im, self._grayscale)
            return im, self._im.info
        
        def _get_meta_data(self, index):
            if not (index is None or index == 0):
                raise IndexError()
            return self._im.info
    
    class Writer(Format.Writer):
        
        def _open(self, **kwargs):
            Image = self.format._init_pillow()
            try:
                self._save_func = Image.SAVE[self.format.plugin_id]
            except KeyError:
                raise RuntimeError('Format %s cannot write images.' %
                                   self.format.name)
            self._fp = self.request.get_file()
            self._meta = {}
            self._meta.update(kwargs)
            self._written = False
        
        def _close(self):
            pass  # request object handled closing _fp
        
        def _append_data(self, im, meta):
            if self._written:
                raise RuntimeError('Format %s only supports single images.' %
                                   self.format.name)
            # Pop unit dimension for grayscale images
            if im.ndim == 3 and im.shape[-1] == 1:
                im = im[:, :, 0]
            self._written = True
            self._meta.update(meta)
            img = ndarray_to_pil(im, self.format.plugin_id)
            if 'bits' in self._meta:
                img = img.quantize()  # Make it a P image, so bits arg is used
            img.save(self._fp, format=self.format.plugin_id, **self._meta)
            img.close()
        
        def set_meta_data(self, meta):
            self._meta.update(meta)


class PNGFormat(PillowFormat):
    """A PNG format based on Pillow.
    
    This format supports grayscale, RGB and RGBA images.
    
    Parameters for reading
    ----------------------
    ignoregamma : bool
        Avoid gamma correction. Default False.
    
    Parameters for saving
    ---------------------
    optimize : bool
        If present and true, instructs the PNG writer to make the output file
        as small as possible. This includes extra processing in order to find
        optimal encoder settings.
    transparency: 
        This option controls what color image to mark as transparent.
    dpi: tuple of two scalars
        The desired dpi in each direction.
    pnginfo: PIL.PngImagePlugin.PngInfo
        Object containing text tags.
    compress_level: int
        ZLIB compression level, a number between 0 and 9: 1 gives best speed,
        9 gives best compression, 0 gives no compression at all. Default is 9.
        When ``optimize`` option is True ``compress_level`` has no effect
        (it is set to 9 regardless of a value passed).
    compression: int
        Compatibility with the freeimage PNG format. If given, it overrides
        compress_level.
    icc_profile:
        The ICC Profile to include in the saved file.
    bits (experimental): int
        This option controls how many bits to store. If omitted,
        the PNG writer uses 8 bits (256 colors).
    quantize: 
        Compatibility with the freeimage PNG format. If given, it overrides
        bits. In this case, given as a number between 1-256.
    dictionary (experimental): dict
        Set the ZLIB encoder dictionary.
    
    """
    
    class Reader(PillowFormat.Reader):
        def _open(self, ignoregamma=False):
            return PillowFormat.Reader._open(self)
        
        def _get_data(self, index):
            im, info = PillowFormat.Reader._get_data(self, index)
            if not self.request.kwargs.get('ignoregamma', False):
                try:
                    gamma = float(info['gamma'])
                except (KeyError, ValueError):
                    pass
                else:
                    scale = float(65536 if im.dtype == np.uint16 else 255)
                    gain = 1.0
                    im = ((im / scale) ** gamma) * scale * gain
            return im, info
                    
    # -- 
    
    class Writer(PillowFormat.Writer):
        def _open(self, compression=None, quantize=None, interlaced=False,
                  **kwargs):
            
            # Better default for compression
            kwargs['compress_level'] = kwargs.get('compress_level', 9)
            
            if compression is not None:
                if compression < 0 or compression > 9:
                    raise ValueError('Invalid PNG compression level: %r' %
                                     compression)
                kwargs['compress_level'] = compression
            if quantize is not None:
                for bits in range(1, 9):
                    if 2**bits == quantize:
                        break
                else:
                    raise ValueError('PNG quantize must be power of two, '
                                     'not %r' % quantize)
                kwargs['bits'] = bits
            if interlaced:
                warn('PIL PNG writer cannot produce interlaced images.')
            
            ok_keys = ('optimize', 'transparency', 'dpi', 'pnginfo', 'bits',
                       'compress_level', 'icc_profile', 'dictionary')
            for key in kwargs:
                if key not in ok_keys:
                    raise TypeError('Invalid arg for PNG writer: %r' % key)
            
            return PillowFormat.Writer._open(self, **kwargs)
        
        def _append_data(self, im, meta):
            if str(im.dtype) == 'uint16' and (im.ndim == 2 or
                                              im.shape[-1] == 1):
                im = image_as_uint(im, bitdepth=16)
            else:
                im = image_as_uint(im, bitdepth=8)
            PillowFormat.Writer._append_data(self, im, meta)


class JPEGFormat(PillowFormat):
    """A JPEG format based on Pillow.
    
    This format supports grayscale, RGB and RGBA images.
    
    Parameters for reading
    ----------------------
    exifrotate : bool
        Automatically rotate the image according to exif flag. Default True.
    
    Parameters for saving
    ---------------------
    quality : scalar
        The compression factor of the saved image (1..100), higher
        numbers result in higher quality but larger file size. Default 75.
    progressive : bool
        Save as a progressive JPEG file (e.g. for images on the web).
        Default False.
    optimize : bool
        On saving, compute optimal Huffman coding tables (can reduce a few
        percent of file size). Default False.
    dpi : tuplw of int
        The pixel density, ``(x,y)``.
    icc_profile : object
        If present and true, the image is stored with the provided ICC profile.
        If this parameter is not provided, the image will be saved with no
        profile attached.
    exif : dict
        If present, the image will be stored with the provided raw EXIF data.
    subsampling : str
        Sets the subsampling for the encoder. See Pillow docs for details.
    qtables : object
        Set the qtables for the encoder. See Pillow docs for details.
    """
    
    class Reader(PillowFormat.Reader):
        def _open(self, exifrotate=True):
            return PillowFormat.Reader._open(self)
        
        def _get_data(self, index):
            im, info = PillowFormat.Reader._get_data(self, index)
            
            # Handle exif
            if 'exif' in info:
                from PIL.ExifTags import TAGS
                info['EXIF_MAIN'] = {}
                for tag, value in self._im._getexif().items():
                    decoded = TAGS.get(tag, tag)
                    info['EXIF_MAIN'][decoded] = value
            
            im = self._rotate(im, info)
            return im, info
        
        def _rotate(self, im, meta):
            """ Use Orientation information from EXIF meta data to 
            orient the image correctly. Similar code as in FreeImage plugin.
            """
            if self.request.kwargs.get('exifrotate', True):
                try:
                    ori = meta['EXIF_MAIN']['Orientation']
                except KeyError:  # pragma: no cover
                    pass  # Orientation not available
                else:  # pragma: no cover - we cannot touch all cases
                    # www.impulseadventure.com/photo/exif-orientation.html
                    if ori in [1, 2]:
                        pass
                    if ori in [3, 4]:
                        im = np.rot90(im, 2)
                    if ori in [5, 6]:
                        im = np.rot90(im, 3)
                    if ori in [7, 8]:
                        im = np.rot90(im)
                    if ori in [2, 4, 5, 7]:  # Flipped cases (rare)
                        im = np.fliplr(im)
            return im
                    
    # -- 
    
    class Writer(PillowFormat.Writer):
        def _open(self, quality=75, progressive=False, optimize=False,
                  **kwargs):
            
            # Check quality - in Pillow it should be no higher than 95
            quality = int(quality)
            if quality < 1 or quality > 100:
                raise ValueError('JPEG quality should be between 1 and 100.')
            quality = min(95, max(1, quality))
            
            kwargs['quality'] = quality
            kwargs['progressive'] = bool(progressive)
            kwargs['optimize'] = bool(progressive)
            
            return PillowFormat.Writer._open(self, **kwargs)
        
        def _append_data(self, im, meta):
            if im.ndim == 3 and im.shape[-1] == 4:
                raise IOError('JPEG does not support alpha channel.')
            im = image_as_uint(im, bitdepth=8)
            PillowFormat.Writer._append_data(self, im, meta)
            return


## Func from skimage

# This cells contains code from scikit-image, in particular from
# http://github.com/scikit-image/scikit-image/blob/master/
# skimage/io/_plugins/pil_plugin.py 
# The scikit-image license applies.


def pil_try_read(im):
    try:
        # this will raise an IOError if the file is not readable
        im.getdata()[0]
    except IOError as e:
        site = "http://pillow.readthedocs.org/en/latest/installation.html"
        site += "#external-libraries"
        pillow_error_message = str(e)
        error_message = ('Could not load "%s" \n'
                         'Reason: "%s"\n'
                         'Please see documentation at: %s'
                         % (im.filename, pillow_error_message, site))
        raise ValueError(error_message)


def _palette_is_grayscale(pil_image):
    if pil_image.mode != 'P':
        return False
    # get palette as an array with R, G, B columns
    palette = np.asarray(pil_image.getpalette()).reshape((256, 3))
    # Not all palette colors are used; unused colors have junk values.
    start, stop = pil_image.getextrema()
    valid_palette = palette[start:stop]
    # Image is grayscale if channel differences (R - G and G - B)
    # are all zero.
    return np.allclose(np.diff(valid_palette), 0)

    
def pil_get_frame(im, grayscale, dtype=None):
    frame = im
    
    if im.format == 'PNG' and im.mode == 'I' and dtype is None:
        dtype = 'uint16'

    if im.mode == 'P':
        if grayscale is None:
            grayscale = _palette_is_grayscale(im)

        if grayscale:
            frame = im.convert('L')
        else:
            
            if im.info.get('transparency', None) is not None:
                # Let Pillow apply the transparency, see issue #210 and #246
                frame = im.convert('RGBA')
            elif im.palette.mode in ('RGB', 'RGBA'):
                # We can do this ourselves. Pillow seems to sometimes screw
                # this up if a  multi-gif has a pallete for each frame ...
                # Create palette array
                p = np.frombuffer(im.palette.getdata()[1], np.uint8)
                # Shape it.
                nchannels = len(im.palette.mode)
                p.shape = -1, nchannels
                if p.shape[1] == 3:
                    p = np.column_stack((p, 255*np.ones(p.shape[0], p.dtype)))
                # Apply palette
                frame_paletted = np.array(im, np.uint8)
                try:
                    frame = p[frame_paletted]
                except Exception:
                    # Ok, let PIL do it. The introduction of the branch that
                    # tests `im.info['transparency']` should make this happen
                    # much less often, but let's keep it, to be safe.
                    frame = im.convert('RGBA')
            else:
                # Let Pillow do it. Unlinke skimage, we always convert
                # to RGBA; palettes can be RGBA.
                if True:  # im.format == 'PNG' and 'transparency' in im.info:
                    frame = im.convert('RGBA')
                else:
                    frame = im.convert('RGB')

    elif im.mode == '1':
        frame = im.convert('L')

    elif 'A' in im.mode:
        frame = im.convert('RGBA')

    elif im.mode == 'CMYK':
        frame = im.convert('RGB')

    if im.mode.startswith('I;16'):
        shape = im.size
        dtype = '>u2' if im.mode.endswith('B') else '<u2'
        if 'S' in im.mode:
            dtype = dtype.replace('u', 'i')
        frame = np.fromstring(frame.tobytes(), dtype)
        frame.shape = shape[::-1]

    else:
        frame = np.array(frame, dtype=dtype)

    return frame


def ndarray_to_pil(arr, format_str=None):
    
    from PIL import Image
    
    if arr.ndim == 3:
        arr = image_as_uint(arr, bitdepth=8)
        mode = {3: 'RGB', 4: 'RGBA'}[arr.shape[2]]

    elif format_str in ['png', 'PNG']:
        mode = 'I;16'
        mode_base = 'I'

        if arr.dtype.kind == 'f':
            arr = image_as_uint(arr)

        elif arr.max() < 256 and arr.min() >= 0:
            arr = arr.astype(np.uint8)
            mode = mode_base = 'L'

        else:
            arr = image_as_uint(arr, bitdepth=16)

    else:
        arr = image_as_uint(arr, bitdepth=8)
        mode = 'L'
        mode_base = 'L'

    array_buffer = arr.tobytes()
    
    if arr.ndim == 2:
        im = Image.new(mode_base, arr.T.shape)
        im.frombytes(array_buffer, 'raw', mode)
    else:
        image_shape = (arr.shape[1], arr.shape[0])
        im = Image.frombytes(mode, image_shape, array_buffer)
    
    return im


## End of code from scikit-image


from .pillowmulti import GIFFormat, TIFFFormat

IGNORE_FORMATS = 'MPEG'

SPECIAL_FORMATS = dict(PNG=PNGFormat, JPEG=JPEGFormat,
                       GIF=GIFFormat, TIFF=TIFFFormat)


def register_pillow_formats():
    
    for id, summary, ext in pillow_formats:
        if id in IGNORE_FORMATS:
            continue
        FormatCls = SPECIAL_FORMATS.get(id, PillowFormat)
        summary = FormatCls._description or summary
        format = FormatCls(id + '-PIL', summary, ext, FormatCls._modes)
        format._plugin_id = id
        if FormatCls is PillowFormat or not FormatCls.__doc__:
            format.__doc__ = pillow_docs[id]
        formats.add_format(format)


register_pillow_formats()



""" 
Helper functions for freezing imageio.
"""

import sys


def get_includes():
    if sys.version_info[0] == 3:
        urllib = ['email', 'urllib.request', ]
    else:
        urllib = ['urllib2']
    return urllib + ['numpy', 'zipfile', 'io']


def get_excludes():
    return []
